import duckdb
from contextlib import contextmanager
import os
import re
# from ezorm import DATABASE
# Context Manager for DuckDB Connection
@contextmanager
def duck_connection(database:str):
    """A context manager to manage DuckDB connections."""
    # Establish the connection
    conn = duckdb.connect(database=database)
    try:
        yield conn  # Yield the connection for usage
    finally:
        conn.close()  # Ensure the connection is closed after use

def remove_escape_characters(text:str)->str:
    pattern = r"[;]"
    return re.sub(pattern, "", text)

def create_directory(db_path:str)->None:
    directory = os.path.dirname(db_path)
    if not os.path.exists(directory):
        os.makedirs(directory)

# def execute(query:str, data:list=[], response=False):
#     with duck_connection(database=DATABASE) as con:
#         records = con.execute(query, data).df()
#     if response==True:
#         return records


from pydantic import BaseModel, Field
from typing import Optional, Type

class EzORMMeta(type(BaseModel)):
    """This is a metaclass to create __table__ that will be used in later step
    Args:
        name (str) : the input class name
        bases (any) : abc
        dct (dict) : this is the annotation of pydantic BaseModel
        table (str) : a specified table name

    Note:
        This will be inspected and debugged later.
    """
    def __new__(cls, name:str, bases, dct:dict, table:str=None):
        new_class = super().__new__(cls, name, bases, dct)
        setattr(new_class, "__table__", table if table else name.lower())
        return new_class

class EzORM(BaseModel, metaclass=EzORMMeta):
    """This is a base class for EzORM"""


# # from ezorm.variables import EzORM
# from ezorm.core import EzORM
# from ezorm.config import DATABASE
# from ezorm.crud import *
# from ezorm.db_management import create_tables, delete_tables
# from ezorm.utils import execute
from ezorm.configuration import settings as ezo


from ezorm.variables import EzORM
from typing import List, Type, Any, Tuple, Dict, Union
from ezorm.utils import remove_escape_characters
import json
from ezorm.configuration import settings
import pandas as pd
from ezorm.validation import isinstance_ezorm


def clean_and_execute(query:str, data:list=[])->pd.DataFrame:
    query = remove_escape_characters(query)
    data = remove_escape_characters(json.dumps(data))
    data = json.loads(data)
    return settings.db(query+";", data)

def get_table_and_schemas(model:Type[EzORM])->Tuple[str, Dict[str, Union[str, int, bool]]]:
    isinstance_ezorm(model)
    table = model.__table__
    schemas = model.model_dump()
    return table, schemas

def get_condition_and_data(schemas:dict)->Tuple[List[str], List[Union[str, int, bool]]]:
    condition = []
    data = []
    for field, value in schemas.items():
        if value is not None:
            condition.append(f"{field}=?")
            data.append(value)
    return condition, data

def Create(model:Type[EzORM])->pd.DataFrame:
    table, schemas = get_table_and_schemas(model)
    query = f"INSERT INTO {table} VALUES ({', '.join(['?' for i in range(len(schemas))])})"
    data = [value for value in schemas.values()]
    return clean_and_execute(query, data)

def Read(model:Type[EzORM])->pd.DataFrame:
    table, schemas = get_table_and_schemas(model)
    where, data = get_condition_and_data(schemas)
    query = f"SELECT * FROM {table}"
    if len(where)>0:
        where = " AND ".join(where)
        query = f"{query} WHERE {where}"
    return clean_and_execute(query, data)

def Update(existing:Type[EzORM], new:Type[EzORM])->pd.DataFrame:
    e_table, e_schemas = get_table_and_schemas(existing)
    n_table, n_schemas = get_table_and_schemas(new)
    if e_table != n_table:
        raise ValueError("table not matched")
    e_where, e_data = get_condition_and_data(e_schemas)
    e_where = " AND ".join(e_where)
    n_set, n_data = get_condition_and_data(n_schemas)
    n_set = ", ".join(n_set)
    query = f"UPDATE {n_table} SET {n_set} WHERE {e_where}"
    data = n_data + e_data
    return clean_and_execute(query, data)

def Delete(model:Type[EzORM])->pd.DataFrame:
    table, schemas = get_table_and_schemas(model)
    where, data = get_condition_and_data(schemas)
    where = " AND ".join(where)
    query = f"DELETE FROM {table} WHERE {where}"
    return clean_and_execute(query, data)


from ezorm.utils import remove_escape_characters
from ezorm.utils import duck_connection
from ezorm.utils import create_directory
# from ezorm import DATABASE
from typing import List, Dict, Any, Type, get_args
from ezorm.variables import EzORM
from ezorm.validation import isinstance_ezorm, issubclass_ezorm
# from ezorm.utils import execute
from ezorm.configuration import settings

def create_tbl_query(table:Type[EzORM])->str:
    issubclass_ezorm(table)
    data_types = {
        str: "TEXT",    # or "TEXT" for unlimited length
        int: "INTEGER",
        float: "FLOAT",    # or "REAL" depending on the database
        bool: "BOOLEAN",
        # "datetime": "DATETIME"  # Use "TIMESTAMP" for PostgreSQL
    }

    query = []

    for field, detail in table.model_fields.items():
        if field != 'table_name':
            proxy = []
            # annotation = detail.annotation
            is_optional = 'Optional' in str(detail.annotation)
            if is_optional:
                dtype, _ = get_args(detail.annotation)
            else:
                dtype = detail.annotation

            proxy.append(f"""{field} {data_types[dtype]}""")
            is_required = not is_optional
            
            if is_required:
                proxy.append(f"""NOT NULL""")
            else:
                default = detail.default
                if (default is not None) and (default != ""):
                    if isinstance(detail.default, bool):
                        proxy.append(f"""DEFAULT {str(default).upper()}""")
                    elif isinstance(detail.default, str):
                        proxy.append(f"""DEFAULT '{default}'""")
                    else:
                        # print("default", type(default), default)
                        proxy.append(f"""DEFAULT {default}""")
                        
            query.append(" ".join(proxy))

    query = ", ".join(query)
    query = remove_escape_characters(query).strip()

    QUERY = f"""CREATE TABLE IF NOT EXISTS {table.__table__} ( {query} );"""
    return QUERY

def delete_tbl_query(table:Type[EzORM])->str:
    issubclass_ezorm(table)
    QUERY = f"""DROP TABLE IF EXISTS {table.__table__};"""
    return QUERY

def create_tables(tables:List[EzORM]):
    create_directory(db_path=settings.database)
    for table in tables:
        query = create_tbl_query(table)
        # print(query)
        # execute(query, [])
        settings.db(query, [])
        print(f"Model: {table.__table__} created successfully")
    print("All tables created successfully")

def delete_tables(tables:List[EzORM]):
    for table in tables:
        query = delete_tbl_query(table)
        # print(query)
        # execute(query, [])
        settings.db(query, [])
        print(f"Model: {table.__table__} deleted successfully")
    print("All tables deleted successfully")


import duckdb
from typing import Any
import pandas as pd
from ezorm.configuration import settings

def duck_engine(query:str, data:list=None, database:str=None)->pd.DataFrame:
    if database is None:
        database = settings.database
    with duckdb.connect(database) as con:
        return con.execute(query, data).df()


from pydantic import BaseModel
from typing import Any

# Define a custom metaclass to handle field registration and validation
class ModelMeta(type(BaseModel)):  # Inherit from the BaseModel's metaclass
    def __new__(cls, name, bases, dct):
        new_class = super().__new__(cls, name, bases, dct)

        # Store fields and their types in the class, using annotations
        fields = {}
        for key, value in dct.get('__annotations__', {}).items():
            # fields[key] = value # this is original code looks like
            fields[key] = key # manually fixing this to key because I want to use the functionality in the ORM class
        new_class._fields = fields
        
        # Also store field types on the class itself
        for field in fields:
            setattr(new_class, field, fields[field])

        return new_class

# Extended BaseModel to handle dynamic behavior for field access
class EzORM(BaseModel, metaclass=ModelMeta):
    __table__:str = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Set the table name to the lowercase name of the class
        cls.__table__ = cls.__name__.lower()
        
    def __init__(self, **kwargs):
        super().__init__(**kwargs)  # Initialize as per Pydantic's logic
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, item: str) -> Any:
        # Return the type of the field if accessed at class level
        if item in self._fields:
            if isinstance(self, type(self)):  # Class-level access
                return self._fields[item]  # Return field type
            else:  # Instance-level access
                return getattr(self, item, None)  # Return the field value
        
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")


from ezorm.variables import EzORM
from typing import Type, Any

def isinstance_ezorm(obj:Type[EzORM]):
    if not isinstance(obj, EzORM):
        raise ValueError(f"Expected a subclass of EzORM, but got {obj}")
    
def issubclass_ezorm(obj:Type[EzORM]):
    if not issubclass(obj, EzORM):
        raise ValueError(f"Expected a subclass of EzORM, but got {obj}")